#include "task.h"

#define NUM_APERIODIC_TASKS 1

#define TASK_SCHEDULER_PRIORITY ( configMAX_PRIORITIES - 1 )
#define TASK_PRIORITY 			( configMAX_PRIORITIES - 2 )

#define USE_TASK_NOTIFY 1

#if ( TIMING_TEST == 2 )
// ============================================================================
extern xType *cs_costs;
// ============================================================================
#endif

/* Periodic task info. TCB for application scheduler. */
struct TaskInfo
{
	TaskHandle_t xHandle;	     /* FreeRTOS task reference. */

	// ----------------------
	ListItem_t xGenericListItem; /* Points to the app scheduled list. */
	ListItem_t xReadyListItem;   /* Points to the scheduler ready list. */

	// ----------------------
	BaseType_t xFinished;        /* 0: running. 1: suspended (called vTaskDelayUntil). */

	// ----------------------
	UBaseType_t xPriority;	     /* Priority. */
	TickType_t xPeriod;		     /* Period. */
	TickType_t xDeadline;        /* Relative deadline. */
	TickType_t xAbsolutDeadline; /* Absolute deadline of the current release. */
	TickType_t xRelease;	     /* Most recent task release absolute time. */

	// ----------------------
	TickType_t xWcet;		     /* Worst case execution time. */
	TickType_t xWcrt;		     /* Worst case response time. */

	// ----------------------
	UBaseType_t uxReleaseCount;  /* Release counter. */

	// ----------------------
	TickType_t xCur; 		     /* Current release tick count. */

	// ----------------------
	void* vExt;                  /* Pointer to a scheduling policy specific structure. */
};

#if defined (__cplusplus)
extern "C" {
#endif

/* Trace blocking and suspended tasks. */
void vSchedulerTaskDelay( void );

void vSchedulerTaskReady( void* xTask );

void vSchedulerTaskBlock( void* xResource );

void vSchedulerTaskSuspend( void* xTask );

/**
 * Application scheduler setup.
 */
void vSchedulerSetup( void );

/**
 * Task scheduler start.
 */
void vSchedulerInit( void );

/**
 * Create a application scheduled task.
 */
BaseType_t xSchedulerTaskCreate( TaskFunction_t pxTaskCode, const char * const pcName, const uint16_t usStackDepth, void * const pvParameters, UBaseType_t uxPriority, struct TaskInfo ** const pxCreatedTask, TickType_t xPeriod, TickType_t xDeadline, TickType_t xWcet );

/**
 * Create a aperiodic application scheduled task.
 */
#define xSchedulerAperiodicTaskCreate( pxTaskCode, pcName, usStackDepth, pvParameters, pxCreatedTask ) xSchedulerTaskCreate( ( pxTaskCode ), ( pcName ), ( usStackDepth ), ( pvParameters ), ( 0 ), ( pxCreatedTask ), ( 0 ), ( 0 ), ( 0 ) )

/**
 * Suspend the caller task until its next period.
 */
void vSchedulerWaitForNextPeriod( void );

/**
 * Release the pxTask from and ISR.
 */
void vSchedulerReleaseFromISR( struct TaskInfo * const pxTask );

/* --- AppSched_Logic -------------------------------------------------- */

void vSchedulerLogicSetup( void );

void vSchedulerTaskSchedulerStartLogic( void );

void vSchedulerLogicAddTask( struct TaskInfo * xTask );

void vSchedulerLogicAddTaskToReadyList( struct TaskInfo *xTask );

void vSchedulerLogicRemoveTaskFromReadyList( struct TaskInfo *xTask );

void vSchedulerTaskSchedulerLogic( struct TaskInfo **xCurrentTask );

struct TaskInfo* xSchedulerGetTaskTCBe( TaskHandle_t xTask );

BaseType_t vSchedulerTaskSchedulerTickLogic( void );

/* --------------------------------------------------------------------- */

/**
 * Hook function for crude debugging.
 */
extern void vSchedulerAppDebug(char a, void * x, void * y);

/* --------------------------------------------------------------------- */

#if defined (__cplusplus)
}
#endif

