/* [[[cog
import cog
import sys
import xml.etree.cElementTree as et

def get_rts_from_element(elem):
    """ extrae el str de elem """

    def get_int(string):
        """ convierte string a int """
        try:
            return int(string)
        except ValueError:
            return int(float(string))

    rts_id, rts = 0, []
    if elem.tag == 'S':
        rts_id = get_int(elem.get("count"))
        for t in elem.iter("i"):
            task = t.attrib
            for k, v in task.items():
                task[k] = get_int(v)
            rts.append(task)

    return rts_id, rts

def test_rts(rts):
    for task in rts:
        task_string = "{0},{1},{2}\n".format(task["C"], task["T"], task["D"])
        sys.stdout.write(task_string)

tick_multiply = int(TICK_MULTIPLY)        
rts_to_test_id = int(RTS_TO_TEST)
rts_to_test = []

context = et.iterparse(RTS_FILE, events=('start', 'end', ))
context = iter(context)
event, root = context.__next__()
for event, elem in context:
    rts_id, rts_to_test = get_rts_from_element(elem)
    if rts_to_test_id == rts_id and rts_to_test and event == "end":
        break
    root.clear()
del context
]]] */
// [[[end]]]
#include "FreeRTOS.h"
#include "task.h"
#include "mbed.h"
#include "scheduler.h"
#include "scheduler_logic.h"
#include "utils.h"
#include "semphr.h"

#if( USE_SLACK == 1 )
#include "slack.h"
#endif

/* [[[cog
for task in rts_to_test:
    cog.outl("#define TASK_{0}_PERIOD {1}".format(task["nro"], int(task["T"])*tick_multiply))
]]] */

// [[[end]]]

/* [[[cog
for task in rts_to_test:
    cog.outl("#define TASK_{0}_WCET {1}".format(task["nro"], int(task["C"])*tick_multiply))
    cog.outl("#define TASK_{0}_RUNTIME {1}".format(task["nro"], int(task["C"])*tick_multiply))
]]] */

// [[[end]]]

#define TASK_APERIODIC_RUNTIME 2000

/* Prototypes for the standard FreeRTOS callback/hook functions implemented
 * within this file. The extern "C" is required to avoid name mangling
 * between C and C++ code.
 */
#if defined (__cplusplus)
extern "C" {
#endif

void vApplicationMallocFailedHook( void );
void vApplicationStackOverflowHook( TaskHandle_t pxTask, char *pcTaskName );

#if( USE_SLACK == 1 )
void vSchedulerNegativeSlackHook( TickType_t xTickCount, BaseType_t xSlack );
#endif

void vSchedulerDeadlineMissHook( struct TaskInfo * xTask, const TickType_t xTickCount );
void vSchedulerStartHook( void );

#if defined (__cplusplus)
}
#endif

static void task_body( void* params );

#if ( USE_SLACK == 1 )
static void aperiodic_task_body( void* params );
#endif

#if ( USE_SLACK == 1 )
static void releaseAperiodicTask1( void );
#endif

/* Create a Serial port, connected to the USBTX/USBRX pins; they represent the
 * pins that route to the interface USB Serial port so you can communicate with
 * a host PC.
 */
static Serial pc( USBTX, USBRX );

static DigitalOut mbed_leds[ ] = { LED1, LED2, LED3, LED4 };

#if ( USE_SLACK == 1 )
InterruptIn button(p8);
int flag = 0;
struct TaskInfo * pxAperiodicTask1;
#endif

#if ( TIMING_TEST == 2 )
    xType *cs_costs;
#endif

int main()
{
	/* === Setup Code ====================================================== */

	/* Set the baud rate of the serial port. */
	pc.baud( 9600 );

#if ( USE_SLACK == 1 )
	/* It is essential that interrupt service routines that make use of the
	 * RTOS API have a logical priority equal to or below that set by the
	 * configMAX_SYSCALL_INTERRUPT_PRIORITY (lower logical priority means
	 * higher numeric value).
	 * See: http://www.freertos.org/RTOS-Cortex-M3-M4.html
	 */
	NVIC_SetPriority( EINT3_IRQn, configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY );

	button.mode(PullDown);
	button.rise( &releaseAperiodicTask1 );
#endif

	/* Set all the mbed leds off. */
	mbed_leds[ 0 ] = 0;
	mbed_leds[ 1 ] = 0;
	mbed_leds[ 2 ] = 0;
    mbed_leds[ 3 ] = 0;

	/* === End of Setup Code =============================================== */
    
#if ( TIMING_TEST == 2 )
    /* Reserve memory for cs_cost[][] array */
    cs_costs = ( xType* ) pvPortMalloc( sizeof( uint32_t ) * ( TASK_COUNT * (RELEASE_COUNT + 1)));

    /* Zeroes cs_cost[][] */
    for(int i = 0; i < TASK_COUNT; i++)
    {
        for(int j = 0; j < RELEASE_COUNT + 1; j++) 
        {            
            (*cs_costs)[i][j] = 0;
        }
    }
#endif

	vSchedulerSetup();

	/* Create the application scheduled tasks. */
	/* [[[cog
    for i, task in enumerate( rts_to_test ):
        cog.outl("xSchedulerTaskCreate( task_body, \"T{0:02}\", 256, NULL, {1}, NULL, TASK_{0}_PERIOD, TASK_{0}_PERIOD, TASK_{0}_WCET );".format(task["nro"], i))
    ]]]*/
    // [[[end]]]

#if ( USE_SLACK == 1 )
	/* Create the application scheduled aperiodic tasks. */
	xSchedulerAperiodicTaskCreate( aperiodic_task_body, "TaskA", 512, NULL, &pxAperiodicTask1 );
#endif

	/* Create and start the scheduler task. */
	vSchedulerInit();

	/* The execution must never reach here. */
}

#if ( USE_SLACK == 1 )
static void releaseAperiodicTask1( void )
{
	if( flag == 0 )
	{
		flag = 1;
		vSchedulerReleaseFromISR( pxAperiodicTask1 );
	}
}
#endif

/**
 * Periodic task body.
 */
static void task_body( void* params )
{
	struct TaskInfo *taskInfo = ( struct TaskInfo * ) params;

#if ( TASK_PRINT_SS == 1 )
	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxBigTaskList );
	ListItem_t * pxAppTasksListItem;
#endif

	for (;;)
	{
#if ( TASK_PRINT == 1 )
		taskENTER_CRITICAL();
#if ( TASK_PRINT_SS == 1 )
		pc.printf( "%d\t\tT%d_START\t%d\t%d\t%d\t", xTaskGetTickCount(), taskInfo->xPriority + 1, taskInfo->uxReleaseCount, taskInfo->xCur, xAvailableSlack );

		pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );
		while( pxAppTasksListEndMarker != pxAppTasksListItem )
		{
			struct TaskInfo_Slack * s = ( struct TaskInfo_Slack * ) ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->vExt;
			pc.printf( "%d\t" , s->xSlack );
			pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
		}
#else
		pc.printf( "%d\t\tT%d_START\t%d\t%d", xTaskGetTickCount(), taskInfo->xPriority + 1, taskInfo->uxReleaseCount, taskInfo->xCur );
#endif
		pc.printf("\n");
		taskEXIT_CRITICAL();
#endif

	    /* [[[cog
        for i, task in enumerate( rts_to_test ):
            cog.outl("if( taskInfo->xPriority == {0} ) {{ vUtilsEatCpu( TASK_{1}_RUNTIME ); }}".format(i, task["nro"]))
        ]]]*/
        // [[[end]]]
		
		if( taskInfo->xPriority == 9 )
		{
#if ( TIMING_TEST == 1 )
			if( task_instance_c >= ( RELEASE_COUNT * TASK_COUNT ) )
			{
				vTaskSuspendAll();
				for(int i = 0; i < TASK_COUNT; i++)
				{
					pc.printf("%d\t", task_instance[i]);
					for(int j = 0; j < RELEASE_COUNT; j++) {
						pc.printf( "%d\t", cs_times[i][j] );
					}
					pc.printf("\n");
				}
				for(;;);
			}
#endif
#if ( TIMING_TEST == 2 )
            if( (*cs_costs)[ ( TASK_COUNT - 1) ][ 0 ] >= RELEASE_COUNT )
			{
				vTaskSuspendAll();
                pc.printf("%d\n", 0);
				for(int i = 0; i < TASK_COUNT; i++)
				{
					pc.printf("%d\t", i);
					/*for(int j = 0; j < RELEASE_COUNT; j++) {
						pc.printf( "%d\t", cs_costs[i][j] );
					}*/
                    for(int j = 1; j < RELEASE_COUNT + 1; j++) {
						pc.printf( "%d\t", (*cs_costs)[i][j]);
					}
					pc.printf("\n");
				}
				for(;;);
			}
#endif
		}

#if ( TASK_PRINT == 1 )
		taskENTER_CRITICAL();
#if ( TASK_PRINT_SS == 1 )
		pc.printf( "%d\t\tT%d_END\t\t%d\t%d\t%d\t", xTaskGetTickCount(), taskInfo->xPriority + 1, taskInfo->uxReleaseCount, taskInfo->xCur, xAvailableSlack );

		pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );
		while( pxAppTasksListEndMarker != pxAppTasksListItem )
		{
			struct TaskInfo_Slack * s = ( struct TaskInfo_Slack * ) ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->vExt;
			pc.printf( "%d\t" , s->xSlack );
			pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
		}
#else
		pc.printf( "%d\t\tT%d_END\t\t%d\t%d", xTaskGetTickCount(), taskInfo->xPriority + 1, taskInfo->uxReleaseCount, taskInfo->xCur );
#endif
		pc.printf("\n");
		taskEXIT_CRITICAL();
#endif

		vSchedulerWaitForNextPeriod();
	}

	/* If the tasks ever leaves the for cycle, kill it. */
	vTaskDelete( NULL );
}

#if ( USE_SLACK == 1 )
static void aperiodic_task_body( void* params )
{
	struct TaskInfo *pxTaskInfo = ( struct TaskInfo * ) params;

	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxBigTaskList );
	ListItem_t * pxAppTasksListItem;

	for (;;)
	{
		taskENTER_CRITICAL();
		pc.printf( "%d\t\tAPERIODIC_START\t%d\t%d\t%d\t", xTaskGetTickCount(), 0, pxTaskInfo->xCur, xAvailableSlack );

		pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );
		while( pxAppTasksListEndMarker != pxAppTasksListItem )
		{
			struct TaskInfo_Slack * s = ( struct TaskInfo_Slack * ) ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->vExt;
			pc.printf( "%d\t" , s->xSlack );
			pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
		}

		pc.printf("\n");
		taskEXIT_CRITICAL();

		vUtilsEatCpu( 2000 );

		taskENTER_CRITICAL();
		pc.printf( "%d\t\tAPERIODIC_END\t%d\t%d\t%d\t", xTaskGetTickCount(), 0, pxTaskInfo->xCur, xAvailableSlack );

		pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );
		while( pxAppTasksListEndMarker != pxAppTasksListItem )
		{
			struct TaskInfo_Slack * s = ( struct TaskInfo_Slack * ) ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->vExt;
			pc.printf( "%d\t" , s->xSlack );
			pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
		}

		pc.printf("\n");
		taskEXIT_CRITICAL();

		/* Reset the flag. */
		flag = 0;

		/* The scheduler will take care of the state of the task. */
		vTaskSuspend( NULL );
	}
}
#endif

void vApplicationMallocFailedHook( void )
{
	taskDISABLE_INTERRUPTS();

	pc.printf( "MALLOC FAILED!\n");

	for( ;; )
	{
	}
}

void vApplicationStackOverflowHook( TaskHandle_t pxTask, char *pcTaskName )
{
	taskDISABLE_INTERRUPTS();

	pc.printf( "STACK OVERFLOW! %s \n", pcTaskName );

	for( ;; )
	{
	}
}

#if ( USE_SLACK == 1 )
void vSchedulerNegativeSlackHook( TickType_t xTickCount, BaseType_t xSlack )
{
    ( void ) xTickCount;
    ( void ) xSlack;
    
	taskDISABLE_INTERRUPTS();

    /*
	pc.printf( "NEGATIVE SLACK !!!\t%d %d - ", xTickCount, xSlack );

	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxBigTaskList );
	ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );

	while( pxAppTasksListEndMarker != pxAppTasksListItem )
	{
		struct TaskInfo_Slack * s = ( struct TaskInfo_Slack * ) ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->vExt;
		pc.printf( "%d\t" , s->xSlack );
		pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
	}

	pc.printf("\n");
    */
    pc.printf("%d\n", 1);

	for( ;; )
	{
	}
}
#endif

extern void vSchedulerDeadlineMissHook( struct TaskInfo * xTask, const TickType_t xTickCount )
{
    ( void ) xTask;
    ( void ) xTickCount;
    
	taskDISABLE_INTERRUPTS();

    pc.printf("%d\n", 1);

	for( ;; )
	{
	}
}

#if ( configUSE_SCHEDULER_START_HOOK == 1 )
#if ( USE_SLACK == 1 )
extern void vSchedulerStartHook()
{
	pc.printf( "\nSetup -- %d\t -- \t%d\t", xTaskGetTickCount(), xAvailableSlack );

	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxBigTaskList );
	ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );

	// Print slack values for the critical instant.
	while( pxAppTasksListEndMarker != pxAppTasksListItem )
	{
		struct TaskInfo_Slack * s = ( struct TaskInfo_Slack * ) ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->vExt;
		pc.printf( "%d\t" , s->xSlack );
		pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
	}

	pc.printf("\nNow, shall we begin? :-) \n");
}
#else
extern void vSchedulerStartHook()
{
	pc.printf("\nNow, shall we begin? :-) \n");
}
#endif
#endif

extern void vSchedulerAppDebug( char a, void *x, void * y)
{
	taskENTER_CRITICAL();
	pc.printf("--- %c - %d - %d\n", a, ( TickType_t ) x, ( TickType_t ) y);
	taskEXIT_CRITICAL();
}
